"""
Copyright (C) 2025 Rockfield Technologies Australia Pty Ltd

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published
by the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
from typing import Mapping, Optional

from .client import MercutoClient
from .exceptions import MercutoClientException, MercutoHTTPException

__all__ = ['MercutoClient', 'MercutoHTTPException', 'MercutoClientException']


def connect(api_key: Optional[str] = None,
            service_token: Optional[str] = None,
            headers: Optional[Mapping[str, str]] = None) -> MercutoClient:
    return MercutoClient().connect(api_key=api_key,
                                   service_token=service_token,
                                   headers=headers)
